@@ -18,6 +18,7 @@ def dict2tuple(name,d):
 
 class Dict2Data:
     _req_keys = ()
+    _subclasses = ()
     """
     - Returns a Data object with dictionary keys as attributes of Data accessible by dot notation or by key. Once an attribute is created, it can not be changed from outside.
     - **Parmeters**
@@ -43,16 +44,19 @@ class Dict2Data:
         if not hasattr(self.__class__,'_req_keys'):
             raise AttributeError("Derived class of `Dict2Data` should have attribute '_req_keys'")
         if isinstance(d,(self.__class__, Dict2Data)):
-            d = d.to_dict() # if nested Dict2Data , must expand here.
+            d = d.to_dict() # if nested Dict2Data , must expand
         # Check if all required keys are present in main level of subclasses
         for key in self.__class__._req_keys:
             if key not in d:
                 raise ValueError(f"Invalid input for {self.__class__.__name__}")
         # ===================
-        for a,b in d.items():
+        for a,b in d.items(): 
             if isinstance(b,(self.__class__, Dict2Data)):
                 b = b.to_dict() # expands self instance !must here.
-            if isinstance(b,(list,tuple)):
+            
+            if a == 'poscar' and 'extra_info' in b:
+                setattr(self,a, PoscarData(b)) # Enables custom methods for PoscarData
+            elif isinstance(b,(list,tuple)):
                 setattr(self,a,[Dict2Data(x) if isinstance(x,dict) else x for x in b])
             else:
                 setattr(self,a,Dict2Data(b) if isinstance(b,dict) else b)
@@ -82,7 +86,7 @@ class Dict2Data:
         """
         result = {}
         for k,v in self.__dict__.items():
-            if isinstance(v,Dict2Data):
+            if isinstance(v,(self.__class__,Dict2Data)):
                 result.update({k:Dict2Data.to_dict(v)})
             else:
                 result.update({k:v})
@@ -106,20 +110,6 @@ class Dict2Data:
         """Creates a namedtuple."""
         return dict2tuple('Data',self.to_dict())
 
-    def write(self,outfile=None):
-        """ If data has an attribute text_plain, writes to file or stdout.
-        - **Parameters**
-            - outfile : Default is None and returns string. If given, writes to file.
-        """
-        if hasattr(self,'text_plain') and self.text_plain:
-            if outfile is None:
-                print(self.text_plain)
-            else:
-                with open(outfile,'w') as f:
-                    f.write(self.text_plain)
-        else:
-            raise AttributeError("No 'text_plain' attribute found.")
-
     def __repr__(self):
         items= []
         for k,v in self.__dict__.items():
@@ -150,6 +140,7 @@ class Dict2Data:
         return self.__dict__[key]
     def items(self):
         return self.__dict__.items()
+
     
 class VasprunData(Dict2Data):
     _req_keys = ('kpath','bands','poscar')
@@ -168,39 +159,26 @@ class SpinData(Dict2Data):
         return super().__repr__().replace("Data","SpinData",1)
     
 class PoscarData(Dict2Data):
-    _req_keys = ('basis','rec_basis','cartesian')
+    _req_keys = ('basis','rec_basis','extra_info')
     def __init__(self,d):
-        if not isinstance(d,dict): # It could be other data type.
-            try:
-                d = d.to_dict()
-            except:
-                raise TypeError(f"{self.__class__} expects dict with valid keys or PoscarData got {d}")
-        fixed_dict = self._add_text_plain(d)
-        super().__init__(fixed_dict)
+        super().__init__(d)
     
     def __repr__(self):
         return super().__repr__().replace("Data","PoscarData",1)
     
-    def _add_text_plain(self, poscar_dict):
-        "Returns poscar_dict with additional attribute `text_plain` after each operation"
-        for key in self.__class__._req_keys:
-            if key not in poscar_dict:
-                raise KeyError(f"Invalid input for PoscarData")
-
-        comment = poscar_dict.get('text_plain',None) or '-- # Exported using pivotpy\n '
-        comment = comment.splitlines()[0].split('#')
-        comment = comment[1].strip() if comment[1:] else ''
-        system = poscar_dict['SYSTEM'].split('#')[0].strip()
-        scale = np.linalg.norm(poscar_dict['basis'][0])
-        unique_d = poscar_dict['unique']
-        out_str = f"{system} # {comment}\n  {scale:<20.14f}\n"
-        out_str += '\n'.join(["{:>22.16f}{:>22.16f}{:>22.16f}".format(*a) for a in poscar_dict['basis']/scale])
-        out_str += "\n  " + '\t'.join(unique_d.keys())
-        out_str += "\n  " + '\t'.join([str(len(v)) for v in unique_d.values()])
-        out_str += "\nCartesian\n" if poscar_dict.get('cartesian',False) else "\nDirect\n"
-        out_str += '\n'.join("{:>21.16f}{:>21.16f}{:>21.16f}".format(*a) for a in poscar_dict['positions'])
-        poscar_dict.update({'text_plain':out_str}) 
-        return poscar_dict
+    @property
+    def coords(self):
+        """Returns the lattice coordinates in cartesian space of the atoms in the poscar data.
+        """
+        from .sio import to_R3 # To avoid circular import
+        return to_R3(self.basis, self.positions)
+    
+    def write(self,sd_list = None, outfile = None, overwrite = False):
+        """Writes the poscar data to a file.
+        """
+        from .sio import write_poscar # To avoid circular import
+        return write_poscar(self,sd_list = sd_list,outfile = outfile,overwrite = overwrite)
+        
 
 class LocpotData(Dict2Data):
     _req_keys = ('ElemName','ElemIndex','SYSTEM')
@@ -285,13 +263,13 @@ def dump(dict_data = None, dump_to = 'pickle',outfile = None,indent=1):
     if dump_to == 'pickle':
         if outfile == None:
             return pickle.dumps(dict_obj)
-        outfile = outfile.split('.')[0] + '.pickle'
+        outfile = os.path.splitext(outfile)[0] + '.pickle'
         with open(outfile,'wb') as f:
             pickle.dump(dict_obj,f)
     if dump_to == 'json':
         if outfile == None:
             return json.dumps(dict_obj,cls = EncodeFromNumpy,indent=indent)
-        outfile = outfile.split('.')[0] + '.json'
+        outfile = os.path.splitext(outfile)[0] + '.json'
         with open(outfile,'w') as f:
             json.dump(dict_obj,f,cls = EncodeFromNumpy,indent=indent)
     return None
